;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Below.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)
;;;
;;;	Note		This entire file is conditional on #+dtp-trace

(in-package "DTP")

;;;----------------------------------------------------------------------------
;;;
;;;	Show-Below

(defgeneric show-below (object &key (answer nil) &allow-other-keys))

;;;----------------------------------------------------------------------------

(defmethod show-below ((object proof) &key (answer nil) &allow-other-keys)
  (declare (ignore answer))
  (dolist (child (objects-below object))
    (output child :parent-id (when *display-query* (dot-id-of object))) ))

;;;----------------------------------------------------------------------------

(defmethod show-below ((object answer) &key (answer nil) &allow-other-keys)
  (declare (ignore answer))
  (output (c-just-conjunction (answer-justification object))
	  :parent-id (when *display-query* (dot-id-of object))
	  :answer (c-just-answer (answer-justification object)) ))

;;;----------------------------------------------------------------------------

(defmethod show-below
    ((object dtp-subgoal)
     &key (answer nil) (parent-id nil) (my-id nil) (record nil)
     (previous-sg-hidden nil) &allow-other-keys )
  (unless *graphic-display* (incf *depth*))
  (unless my-id (setq my-id (dot-id-of object)))
  (if answer
      (when (typep answer 'answer)
	(let ((just (answer-justification answer))
	      (hidden
	       (and (hidden-subgoal parent-id) (not previous-sg-hidden)) )
	      (label nil) )
	  ;; (when (typep just 'c-justification)
	  ;; (setq label
	  ;; (binding-label-between object (c-just-conjunction just) answer) ))
	  (output just :link-label label :record (when hidden record)
		  :previous-sg-hidden hidden
		  :parent-id (if hidden parent-id my-id) )))
    (unless (find object *explored*)
      (push object *explored*)
      (dolist (child (objects-below object))
	(output child :parent-id my-id
		:link-label (when (typep child 'dtp-conjunction)
			      (binding-label-between object child) )))))
  (unless *graphic-display* (decf *depth*)) )

;;;----------------------------------------------------------------------------

(defmethod show-below
    ((object dtp-conjunction)
     &key (answer nil) (parent-id nil) (my-id nil) (record nil)
     (link-label nil) (previous-sg-hidden nil) &allow-other-keys )
  (unless (hidden-conjunction object)
    (if my-id
	(setq parent-id my-id)
      (setq parent-id (dot-id-of object)) )
    (setq link-label nil) )
  (if answer
      (when (typep answer 'answer)
	(loop
	    with sg-justs =
	      (remove-if-not #'s-justification-p (answer-justification answer))
	    for just in sg-justs
	    for o-record =
	      (if (hidden-conjunction object)
		  record
		(s-just-conjunct-number just) )
	    for o-hidden =
	      (if (hidden-conjunction object)
		  previous-sg-hidden
		nil )
	    do (output (s-just-subgoal just) :parent-id parent-id
		       :record o-record :previous-sg-hidden o-hidden
		       :link-label link-label :answer (s-just-answer just) )
	       ))
    (dolist (child (objects-below object))
      (output child :parent-id parent-id :link-label link-label) )
    ))

;;;----------------------------------------------------------------------------

(defmethod show-below
    ((object dtp-conjunct)
     &key (parent-id nil) (link-label nil) (answer nil) &allow-other-keys )
  (declare (ignore answer))
  (let ((parent (dot-id-to-object parent-id)))
    (if (typep parent 'dtp-conjunction)
	(let ((real-parent (slot-value object 'parent-conjunction))
	      num )
	  ;; Graph parent <> real parent for forward inference conjunctions
	  (if (eq parent real-parent)
	      (setq num (position object (slot-value parent 'list)))
	    (setq num
	      (+ (position object (slot-value real-parent 'list))
		 (- (length (slot-value parent 'list))
		    (length (slot-value real-parent 'list)) ))))
	  (dolist (child (objects-below object))
	    (output
	     child :parent-id parent-id :record num :link-label link-label )))
      (dolist (child (objects-below object))
	(output child :parent-id parent-id :link-label link-label) ))))

;;;----------------------------------------------------------------------------

(defmethod show-below
    ((object c-justification)
     &key (parent-id nil) (answer nil) (record nil) (link-label nil)
     (previous-sg-hidden nil) &allow-other-keys )
  (declare (ignore answer))
  (output (c-just-conjunction object) :parent-id parent-id :record record
	  :link-label link-label :answer (c-just-answer object)
	  :previous-sg-hidden previous-sg-hidden ))

;;;----------------------------------------------------------------------------

(defmethod show-below
    ((object justification) &key (answer nil) &allow-other-keys)
  "Nothing is below a justification"
  (declare (ignore answer)) )

;;;----------------------------------------------------------------------------
;;;
;;;	Objects-Below

(defgeneric objects-below (object))

;;;----------------------------------------------------------------------------

(defmethod objects-below ((object proof))
  (remove-single-conjunctions
   (remove-uninitialized
    (append
     (proof-used-conjunctions object)
     (proof-query-conjunctions object) ))))

;;;----------------------------------------------------------------------------

(defmethod objects-below ((object dtp-subgoal))
  (remove
   nil
   (append
    (when (listp (slot-value object 'used-inferences)) ; Unless :UNINITIALIZED
      (slot-value object 'used-inferences) )
    (when (listp (slot-value object 'inferences)) ; Unless :UNINITIALIZED
      (slot-value object 'inferences) ))))

;;;----------------------------------------------------------------------------

(defmethod objects-below ((object dtp-conjunction))
  "Either conjuncts below, or (if singleton) subgoals below"
  (let (conjuncts)
    (loop
	for instance in (slot-value object 'instance-conjunctions)
	append (slot-value instance 'list)
	into i-conjuncts
	finally
	  (setq conjuncts (append (slot-value object 'list) i-conjuncts)) )
    (setq conjuncts (remove-uninitialized conjuncts))
    (if (eq (length conjuncts) 1)
	(objects-below (first conjuncts))
      conjuncts )))

;;;----------------------------------------------------------------------------

(defmethod objects-below ((object dtp-conjunct))
  (remove-duplicates
   (remove-uninitialized
    (append
     (slot-value object 'used-subgoals)
     (list (slot-value object 'subgoal)) ))))

;;;----------------------------------------------------------------------------

(defun remove-single-conjunctions (conjunctions)
  #+dtp-type (declare (type list conjunctions))
  (reduce
   #'append
   (mapcar
    #'(lambda (c)
	#+dtp-types (declare (type dtp-conjunction c))
	(if (eq (length (slot-value c 'list)) 1)
	    (objects-below (first (slot-value c 'list)))
	  (list c) ))
    conjunctions )))

;;;----------------------------------------------------------------------------

(defun remove-uninitialized (l)
  #+dtp-types (declare (type list l))
  (remove-if-not #'(lambda (x) (typep x 'dtp-object)) l) )

;;;----------------------------------------------------------------------------
