;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Classes.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

;;;----------------------------------------------------------------------------

(defclass dtp-object () ()
  (:documentation "Superclass of all DTP objects") )

;;;----------------------------------------------------------------------------

(defclass dtp-proof-node (dtp-object) ()
  (:documentation "Node in the proof space") )

;;;----------------------------------------------------------------------------

(defclass dtp-subgoal (dtp-proof-node)
	  ((literal
	    :initarg :literal
	    :initform nil
	    :type (or literal-node null) )
	   (depth
	    :initarg :depth
	    :initform nil
	    :type (or null (integer 0 *))
	    :documentation "Length of minimal path from root to this node" )
	   (answers
	    :initform nil
	    :type list )
	   (reduction-answers
	    :initform nil
	    :type list )
	   (conjuncts-to-propogate-to
	    :initarg :propogate-to
	    :initform nil
	    :type list )
	   (remaining-ancestor-subgoals
	    :initform :uninitialized
	    #-allegro-v4.1 :type #-allegro-v4.1 (or list (eql :uninitialized))
	    :documentation "Used for reduction" )
	   (assumables
	    :initform :uninitialized
	    #-allegro-v4.1 :type #-allegro-v4.1 (or list (eql :uninitialized))
	    :documentation "Used for residue" )
	   (inferences
	    :initform :uninitialized
	    #-allegro-v4.1 :type #-allegro-v4.1 (or list (eql :uninitialized))
	    :documentation "List of conjunction nodes" )
           (forward-inference-conjunctions
            :initform nil
            :type list
            :documentation "Conjunction nodes that are waiting for propogation"
            )
	   (parent-subgoal
	    :initform nil
	    :type (or null dtp-subgoal)
	    :documentation "Upward pointer in proof space" )
	   #+dtp-trace
	   (used-inferences
	    :initform nil
	    :type list
	    :documentation "Only used for tracing" )
	   #+dtp-trace
	   (used-conjuncts
	    :initform nil
	    :type list
	    :documentation "Only used for tracing, from -to-propogate-to" )
	   #+dtp-trace
	   (used-ancestors
	    :initform nil
	    :type list
	    :documentation "Only used for tracing" )
	   )
  (:documentation "Subgoal node in proof space") )

(defmethod print-object ((object dtp-subgoal) stream)
  (with-slots (literal answers inferences) object
    (format stream "#<DTP Subgoal ")
    (if (literal-node-p literal)
	(print-literal-node literal :s stream)
      (format stream "?") )
    (format stream " with ~D answer~:P" (length answers))
    (when (listp inferences)
      (if inferences
	  (format stream " [~D task~:P pending]" (length inferences))
	(format stream " [complete]") ))
    (format stream ">") ))

;;;----------------------------------------------------------------------------

(defclass dtp-conjunction (dtp-proof-node)
	  ((list
	    :initarg :list
	    :initform nil
	    :type list
	    :documentation "List of conjuncts" )
	   (stack :initform nil)
	   (stack-pointer
	    :initform 0
	    :type (integer -1 *)
	    :documentation "Number of current conjunct" )
	   (backtrack-pointer
	    :initform -1
	    :type (integer -1 *)
	    :documentation
	    "Conjuncts from 0 to here must not be backjumped over" )
	   (parent-subgoal
	    :initarg :parent
	    :initform nil
	    :type (or dtp-subgoal null)
	    :documentation "Parent of NIL means this is a query-conjunction" )
	   (binding-list
	    :initarg :binding-list
	    :initform nil
	    :type binding-list
	    :documentation "From inference" )
	   (label
	    :initarg :label
	    :initform nil
	    :type (or label null)
	    :documentation "From inference" )
	   (residue
	    :initarg :residue
	    :initform nil
	    :type list
	    :documentation "Only when summarizing for instance conjunction" )
	   (ae-binding-list
	    :initarg :ae-binding-list
	    :initform nil
	    :type binding-list
	    :documentation
	    "Needed for disjunctive answers via answer extraction" )
	   #+dtp-trace
	   (origin
	    :initarg :origin
	    :type symbol
	    :documentation "KB node ID of parent rule, used for tracing" )
	   #+dtp-trace
	   (answers
	    :initform nil
	    :type list
	    :documentation "List of answers found by the conjunction" )
	   #+dtp-trace
	   (instance-conjunctions
	    :initform nil
	    :type list
	    :documentation "Only used for tracing" )
	   )
  (:documentation "Conjunction (from inference)") )

(defmethod print-object ((object dtp-conjunction) stream)
  #+dtp-types (declare (type stream stream))
  (with-slots (list stack-pointer) object
    (format stream "#<DTP Conjunction")
    (when list
      (format stream " [~D]:" stack-pointer)
      (dolist (conjunct list)
	(when (typep conjunct 'dtp-conjunct)
	  (with-slots (literal answer-count) conjunct
	    (format stream " ")
	    (if (literal-node-p literal)
		(print-literal-node literal :s stream)
	      (format stream "?") )
	    (format stream "/~D" answer-count) ))))
    (format stream ">") ))

;;;----------------------------------------------------------------------------

(defclass dtp-conjunct (dtp-proof-node)
	  ((literal
	    :initarg :literal
	    :initform nil
	    :type (or literal-node null) )
	   (parent-conjunction
	    :initarg :parent
	    :type dtp-conjunction
	    :documentation "Must be initialized when first created" )
	   (binding-list
	    :initform nil
	    :type list
	    :documentation "Apply to literal, then search for subgoal" )
	   (transform-binding-list
	    :initform nil
	    :type list
	    :documentation "Apply to answers from subgoal before valid" )
	   (answer-count
	    :initform 0
	    :type (integer 0 *) )
	   (ranswer-count
	    :initform 0
	    :type (integer 0 *) )
	   (subgoal
	    :initform :uninitialized )
	   (nogoods
	    :initform :uninitialized
	    #-allegro-v4.1 :type
	    #-allegro-v4.1 (or list (eql :uninitialized))
	    :documentation "Backjumping dependency analysis" )
	   #+dtp-trace
	   (used-subgoals
	    :initform nil
	    :type list
	    :documentation "For tracing only" ))
  (:documentation "Conjunct") )

(defmethod print-object ((object dtp-conjunct) stream)
  (with-slots (literal answer-count) object
    (format stream "#<DTP Conjunct ")
    (if (literal-node-p literal)
	(print-literal-node literal :s stream)
      (format stream "?") )
    (format stream " with ~D answer~:P>" answer-count) ))

;;;----------------------------------------------------------------------------
