;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Conjunctions.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

;;;----------------------------------------------------------------------------
;;;
;;;	Public

;;;----------------------------------------------------------------------------

(defmethod expand ((conjunction dtp-conjunction))
  (with-slots
      (list stack stack-pointer parent-subgoal binding-list label answers)
      conjunction
    (unless (< stack-pointer 0)
      (let ((conjunct (nth stack-pointer list)))
	(if conjunct
	    (expand conjunct)
	  (let (answer)
	    (setq answer (merge-conjunction-answers conjunction))
	    (unless (eq answer :not-an-answer)
	      #+dtp-trace
	      (let ((propogate-answer (copy-answer answer)))
		(add-to-end answer answers)
		(setf (answer-justification propogate-answer)
		  (make-c-justification
		   :conjunction conjunction
		   :answer answer ))
		(setq answer propogate-answer) )
	      (note-solution conjunction)
	      (if parent-subgoal
		  (propogate answer parent-subgoal)
		(propogate answer *proof*) ))
	    (pop stack)
	    (decf stack-pointer) ))
	))))

#+dtp-trace
(defmethod expand :around ((conjunction dtp-conjunction))
  "For proof tracing"
  (when (find :proofs *trace*)
    (indent-line)
    (format *debug-io* "Expanding conjunction ")
    (debug-print conjunction)
    (format *debug-io* "~%") )
  (call-next-method) )

;;;----------------------------------------------------------------------------

(defmethod exhausted-p ((conjunction dtp-conjunction))
  (< (slot-value conjunction 'stack-pointer) 0) )

;;;----------------------------------------------------------------------------

(defmethod propogate (answer (conjunction dtp-conjunction))
  (with-slots (list stack stack-pointer) conjunction
    (if (eq answer :not-an-answer)
	(backtrack conjunction)
      (let ((next-conjunct (nth (incf stack-pointer) list))
	    mgu )
	(push answer stack)
	(setq mgu (merge-conjunction-binding-lists conjunction))
	(when next-conjunct
	  (if (eq mgu :not-a-binding-list)
	      (progn
		(pop stack)
		(decf stack-pointer)
		(return-from propogate) )
	    (setf (slot-value next-conjunct 'binding-list) mgu) ))
	(expand conjunction) ))
      ))

#+dtp-trace
(defmethod propogate :around (answer (conjunction dtp-conjunction))
  "For proof tracing"
  #+dtp-types (declare (type answer answer))
  (when (find :proofs *trace*)
    (indent-line)
    (format *debug-io* "Propogating ~S to conjunction " answer)
    (debug-print conjunction)
    (format *debug-io* "~%") )
  (call-next-method) )

;;;----------------------------------------------------------------------------
;;;
;;;	Private

;;;----------------------------------------------------------------------------

(defun merge-conjunction-answers (conjunction)
  "Returns merge of answers (plus conj bl), or :NOT-AN-ANSWER"
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let ((answer (merge-answers (slot-value conjunction 'stack))))
    (when (eq answer :not-an-answer)
      (return-from merge-conjunction-answers :not-an-answer) )
    (when (slot-value conjunction 'binding-list)
      (setq answer
	(nanswer-merge-binding-list
	 answer (slot-value conjunction 'binding-list) ))
      (when (eq answer :not-an-answer)
	(return-from merge-conjunction-answers :not-an-answer) ))
    (when (slot-value conjunction 'label)
      (setq answer
	(nanswer-merge-label answer (slot-value conjunction 'label)) ))
    (when (slot-value conjunction 'residue)
      (setq answer
	(nanswer-merge-residue answer (slot-value conjunction 'residue)) )
      (when (eq answer :not-an-answer)
	(return-from merge-conjunction-answers :not-an-answer) ))
    (when (slot-value conjunction 'ae-binding-list)
      (push (slot-value conjunction 'ae-binding-list)
	    (answer-ae-binding-lists answer) ))
    #+dtp-trace
    (record-justification conjunction answer)
    #+dtp-trace
    (let ((origin (slot-value conjunction 'origin)))
      (unless (eq origin 'query)
	(push (make-l-justification :id origin)
	      (answer-justification answer) )))
    answer ))

;;;----------------------------------------------------------------------------

(defun merge-conjunction-binding-lists (conjunction)
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let ((bls (mapcar #'answer-binding-list (slot-value conjunction 'stack))))
    (setq bls (cons (slot-value conjunction 'binding-list) bls))
    (setq bls (remove nil bls))
    (if (cdr bls)
	(let ((merge-bl (merge-binding-lists bls)))
	  (if merge-bl merge-bl :not-a-binding-list) )
      (first bls) )))

;;;----------------------------------------------------------------------------

(defun merge-conjunction-labels (conjunction)
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let ((labels (mapcar #'answer-label (slot-value conjunction 'stack))))
    (setq labels (cons (slot-value conjunction 'label) labels))
    (setq labels (remove nil labels))
    (if (cdr labels)
	(reduce #'label-and labels)
      (first labels) )))

;;;----------------------------------------------------------------------------

(defun merge-conjunction-residues (conjunction)
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let ((residues (mapcar #'answer-residue (slot-value conjunction 'stack))))
    (setq residues (cons (slot-value conjunction 'residue) residues))
    (setq residues (remove nil residues))
    (if (cdr residues)
	(let ((merge-res (reduce #'residue-merge residues)))
	  (if (eq merge-res :not-a-residue) :not-a-residue merge-res) )
      (first residues) )))

;;;----------------------------------------------------------------------------

(defun list-to-conjunction (list)
  #+dtp-types (declare (type list list))
  (loop
      for sublist in list
      collect (make-instance 'dtp-conjunct :literal (list-to-literal sublist))
      into conjuncts
      finally
	(let (conjunction)
	  (setq conjunction
	    (make-instance 'dtp-conjunction :list conjuncts
			   #+dtp-trace :origin #+dtp-trace 'query) )
	  (dolist (conjunct conjuncts)
	    (setf (slot-value conjunct 'parent-conjunction) conjunction) )
	  (return conjunction) )))

;;;----------------------------------------------------------------------------

(defun make-instance-conjunction (conjunction)
  "Return a new forward-chaining conjunction using CONJUNCTION state"
  (let ((bl (merge-conjunction-binding-lists conjunction))
        (label (merge-conjunction-labels conjunction))
	(residue (merge-conjunction-residues conjunction))
        new-conj conjuncts )
    (when (eq bl :not-a-binding-list)
      (return-from make-instance-conjunction) )
    (with-slots (list parent-subgoal stack-pointer #+dtp-trace origin)
	conjunction
      (setq new-conj
	(make-instance 'dtp-conjunction
	  :parent parent-subgoal #+dtp-trace :origin #+dtp-trace origin
	  :binding-list bl :label label :residue residue ))
      (setq conjuncts
	(mapcar
	 #'(lambda (conjunct)
	     (setq conjunct (copy-conjunct conjunct))
	     (setf (slot-value conjunct 'parent-conjunction) new-conj)
	     #+dtp-trace (setf (slot-value conjunct 'used-subgoals) nil)
	     conjunct )
	 (nthcdr stack-pointer list) ))
      (setf (slot-value new-conj 'list) conjuncts)
      new-conj )))

;;;----------------------------------------------------------------------------

(defun copy-conjunct (conjunct)
  "Return a copy of the conjunct"
  (let (nc)
    (with-slots (literal parent-conjunction binding-list transform-binding-list
		 answer-count subgoal #+dtp-trace used-subgoals )
	conjunct
      (setq nc
	(make-instance 'dtp-conjunct
	  :literal literal :parent parent-conjunction ))
      (setf (slot-value nc 'binding-list) binding-list
	    (slot-value nc 'transform-binding-list) transform-binding-list
	    (slot-value nc 'answer-count) answer-count
	    (slot-value nc 'subgoal) subgoal )
      #+dtp-trace (setf (slot-value nc 'used-subgoals) used-subgoals) )
    nc ))

;;;----------------------------------------------------------------------------

(defun debug-print (conjunction)
  "Concise summary of object"
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (loop
      for remaining on (slot-value conjunction 'list)
      for literal = (slot-value (first remaining) 'literal)
      for in-middle = nil then t
      do (when in-middle (format *debug-io* "^"))
	 (print-literal-node literal :s *debug-io*) ))

;;;----------------------------------------------------------------------------

(defun note-solution (conjunction)
  "Solution found, so must backtrack instead of backjump"
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (with-slots (list backtrack-pointer) conjunction
    (setq backtrack-pointer (1- (length list))) ))

;;;----------------------------------------------------------------------------

#+dtp-trace
(defun record-justification (conjunction answer)
  #+dtp-type (declare (type dtp-conjunction conjunction))
  #+dtp-type (declare (type answer answer))
  (loop
      for conjunct in (slot-value conjunction 'list)
      for c-answer in (reverse (slot-value conjunction 'stack))
      for conj-num from 0
      collect
	(with-slots
	    (transform-binding-list answer-count ranswer-count subgoal)
	    conjunct
	  (let ((count (1- (- answer-count ranswer-count)))
		ans )
	    (cond
	     ((typep c-answer 'reduction-answer)
	      (setq ans c-answer) )
	     ((>= count 0)
	      (setq ans (nth count (slot-value subgoal 'answers))) )
	     (t
	      (setq ans :unknown) ))
	    (make-s-justification
	     :conjunct-number conj-num :subgoal subgoal
	     :answer ans :t-bl transform-binding-list )))
      into justs
      finally (setf (answer-justification answer) justs) ))

;;;----------------------------------------------------------------------------
