;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Dotify.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)
;;;
;;;	Note		This entire file is conditional on #+dtp-trace

(in-package "DTP")

;;;----------------------------------------------------------------------------

(defgeneric dotify
    (object &key (parent-id nil) (answer nil) &allow-other-keys) )

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object proof) &key (parent-id nil) (answer nil) &allow-other-keys)
  (declare (ignore parent-id answer))
  (when *display-query*
    (format *show-stream* "  ~A [label=~S] ;~%"
	    (dot-node-name object) (dot-name-of object) )))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object answer) &key (parent-id nil) (answer nil) &allow-other-keys)
  (declare (ignore parent-id answer))
  (when *display-query*
    (format *show-stream* "  ~A [label=~S] ;~%"
	    (dot-node-name object) (dot-name-of object) )))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object dtp-subgoal)
     &key (answer nil) (parent-id nil) (my-id nil) (record nil)
     (link-label nil) (previous-sg-hidden nil) &allow-other-keys )
  
  (let ((my-parent nil))
    
    (when answer (setq link-label nil))
    (when parent-id (setq my-parent (dot-id-to-object parent-id)))
    (unless my-id (setq my-id (dot-id-of object)))
    
    (when (and answer (hidden-subgoal parent-id) (not previous-sg-hidden))
      (return-from dotify) )

    ;; Incoming arcs
    (when parent-id
      (if record
	  (format *show-stream* "  ~A:c~D -> ~A"
		  (dot-node-name parent-id) record (dot-node-name my-id) )
	(format *show-stream* "  ~A -> ~A"
		(dot-node-name parent-id) (dot-node-name my-id) ))
      (let (slave-link)
	(setq slave-link
	  (and (not answer)
	       my-parent
	       (not (eq (subgoal-parent-of my-parent)
			(slot-value object 'parent-subgoal) ))))
	(when (or slave-link link-label)
	  (format *show-stream* " [")
	  (when slave-link
	    (format *show-stream* "weight=1,")
	    (if *display-color*
		(format *show-stream* "color=blue")
	      (format *show-stream* "style=dotted") ))
	  (when (and slave-link link-label) (format *show-stream* ","))
	  (when link-label (format *show-stream* "label=\"~A\"" link-label))
	  (format *show-stream* "]") ))
      (format *show-stream* " ;~%") )
  
    ;; Outgoing arcs
    (unless (and (find object *explored*) (not answer))
      (format *show-stream* "  ~A [shape=plaintext,label=~S] ;~%"
	      (dot-node-name my-id) (dot-name-of object answer) )
      (unless answer (dotify-reductions object my-id)) )
    ))

(defun dotify-reductions (subgoal parent-id)
  #+dtp-types (declare (type dtp-subgoal subgoal))
  #+dtp-types (declare (type integer parent-id))
  (dolist (r-ans (slot-value subgoal 'reduction-answers))
    (when (eq subgoal (r-just-leaf-subgoal (ranswer-justification r-ans)))
      (dotify r-ans :parent-id parent-id) )))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object dtp-conjunction)
     &key (answer nil) (parent-id nil) (my-id nil) (record nil)
     (link-label nil) &allow-other-keys )

  (when (hidden-conjunction object)
    (return-from dotify) )	; Hide intermediate node in full space
  
  (let ((name (dot-name-of object answer)))
    
    (when (and answer (eq name :lookup))
      (return-from dotify) )	; Hide LOOKUP nodes in explanations
    
    (unless my-id (setq my-id (dot-id-of object)))
    (when answer (setq link-label nil))
    
    ;; Incoming arcs
    (when parent-id
      (if record
	  (format *show-stream* "  ~A:c~D -> ~A"
		  (dot-node-name parent-id) record (dot-node-name my-id) )
	(format *show-stream* "  ~A -> ~A"
		(dot-node-name parent-id) (dot-node-name my-id) ))
      (when link-label
	(format *show-stream* " [label=\"~A\"]" link-label) )
      (format *show-stream* " ;~%") )
    
    ;; Conjunction
    (format *show-stream* "  ~A " (dot-node-name my-id))
    (if (eq name :lookup)
	(format *show-stream*
		"[shape=record,fontsize=9,label=\"Lookup\"] ;~%" )
      (format *show-stream* "[shape=record,label=~S] ;~%" name) )
    ))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object dtp-conjunct) &key (parent-id nil) (answer nil) &allow-other-keys)
  "No DOT output for a conjunct"
  (declare (ignore parent-id answer)) )

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object reduction-answer)
     &key (parent-id nil) (answer nil) &allow-other-keys )
  (declare (ignore answer))
  (let ((id (dot-id-of object :unique t)))
    (format *show-stream*
	    "  ~A [shape=record,fontsize=9,label=\"Reduction\"] ;~%"
	    (dot-node-name id) )
    (format *show-stream* "  ~A -> ~A [label=\"~A\"] ;~%"
	    (dot-node-name parent-id) (dot-node-name id)
	    (dot-name-of-bl (ranswer-binding-list object)) )))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object l-justification)
     &key (parent-id nil) (answer nil) (link-label nil) &allow-other-keys )
  (declare (ignore answer))
  (return-from dotify)		; Don't output LOOKUP nodes
  (when parent-id
    (format *show-stream* "  ~A -> ~A"
	    (dot-node-name parent-id) (dot-node-name object) )
    (when link-label
      (format *show-stream* " [label=\"~A\"]" link-label) )
    (format *show-stream* " ;~%") )
  (format *show-stream* "  ~A " (dot-node-name object))
  (format *show-stream* "[shape=record,fontsize=9,label=\"Lookup\"] ;~%") )

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object r-justification)
     &key (parent-id nil) (answer nil) (link-label nil) &allow-other-keys )
  (declare (ignore answer))
  (let ((id (dot-id-of object :unique t)))
    (when parent-id
      (format *show-stream* "  ~A -> ~A"
	      (dot-node-name parent-id) (dot-node-name id) )
      (when link-label
	(format *show-stream* " [label=\"~A\"]" link-label) )
      (format *show-stream* " ;~%") )
    (format *show-stream* "  ~A " (dot-node-name id))
    (format *show-stream* "[shape=record,fontsize=9,label=\"Reduction\"] ;~%")
    ))

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object res-justification)
     &key (parent-id nil) (answer nil) (link-label nil) &allow-other-keys )
  (declare (ignore answer))
  (when parent-id
    (format *show-stream* "  ~A -> ~A"
	    (dot-node-name parent-id) (dot-node-name object) )
    (when link-label
      (format *show-stream* " [label=\"~A\"]" link-label) )
    (format *show-stream* " ;~%") )
  (format *show-stream* "  ~A " (dot-node-name object))
  (format *show-stream* "[shape=record,fontsize=9,label=\"Residue\"] ;~%") )

;;;----------------------------------------------------------------------------

(defmethod dotify
    ((object justification)
     &key (parent-id nil) (answer nil) &allow-other-keys )
  "No DOT output for an unknown justification"
  (declare (ignore parent-id answer)) )

;;;----------------------------------------------------------------------------
