;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		DTP.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)
;;;
;;;	Version history
;;;		1.0	Resolution-based inference system
;;;		2.0	Subgoaling inference system
;;;		2.1	Corrected reduction inference combined with caching
;;;		2.2	Depth limits and iteration (Broken implementation)
;;;		2.3	Reduction check at conjunct instead of subgoal
;;;		2.4	General "view" tool on proofs and answers
;;;
;;;	Documentation Notes
;;;
;;;	Exported functions are intended to be called by users.  All other
;;;	functions and symbols are intended to be internal.
;;;
;;;	Many files have a "public" and a "private" section.  Functions defined
;;;	in the "private" section are ONLY called by functions within the same
;;;	file.  All functions called by routines in other files are placed in
;;;	the "public" section.

(in-package "COMMON-LISP-USER")

(eval-when (compile eval load)
  (defpackage "DTP"
    #-lucid (:use "COMMON-LISP") ))

(in-package "DTP")

;;;----------------------------------------------------------------------------
;;;
;;;	Options to change for new system installation

(eval-when (compile load eval)
  
  ;; Customize feature
  ;; Controls: Default directories for files.  If present, set up for Geddis
  (pushnew :customize-dfg *features*)
  
  ;; Debug feature
  ;; Controls: Whether to optimize compilation steps, or leave debugging info 
  (pushnew :dtp-debug *features*)
  
  ;; Trace feature
  ;; Controls: Whether to include code and data structures for watching the
  ;;    inference in the middle of problem solving (and for examining
  ;;    proof spaces afterwards).  Most applications will want this option
  ;;    present, but some (e.g. an autonomous process) might wish the extra
  ;;    speed and smaller space, sacrificing user-friendlyness.
  (pushnew :dtp-trace *features*)
  
  ;; Type feature
  ;; Controls: Whether to declare data types of variables.  Some lisps can't
  ;;    handle this.
  (pushnew :dtp-types *features*)
  )

#-(or lucid allegro-v4.1)
(eval-when (compile load eval)
  (setf (logical-pathname-translations "dtp")
    #+(and customize-dfg allegro)
    '(("*.*.*" "~geddis/Lisp/DTP/")
      ("src;*.*.*" "~geddis/Lisp/DTP/Source/")
      ("bin;*.*.*" "~geddis/Lisp/DTP/Compiled/")
      ("doc;*.*.*" "~geddis/Lisp/DTP/Documentation/")
      ("logic;*.*.*" "~geddis/Lisp/DTP/Logic/") )
    #+(and customize-dfg mcl)		; Bug means most specific must be first
    '(("src;*.*.*" "Don Geddis:DTP:")
      ("bin;*.*.*" "Don Geddis:DTP:")
      ("doc;*.*.*" "Don Geddis:DTP:")
      ("logic;*.*.*" "Don Geddis:DTP:")
      ("*.*.*" "Don Geddis:DTP:") )
    #-customize-dfg  
    '(("*.*.*" "")
      ("src;*.*.*" "")
      ("bin;*.*.*" "")
      ("doc;*.*.*" "")
      ("logic;*.*.*" "") )))

;;;----------------------------------------------------------------------------

(eval-when (compile load eval)
  (defparameter *dtp-major-version* 2)
  (defparameter *dtp-minor-version* 4)
  (defparameter *dtp-compilation-status*
      #+dtp-debug "debugging"
      #-dtp-debug "production" )
  (defparameter *dtp-tracing-status*
      #+dtp-trace "tracing"
      #-dtp-trace "no tracing" )
  (defparameter *dtp-typing-status*
      #+dtp-types "types"
      #-dtp-types "no types" )
  )

(eval-when (compile load eval)
  (defparameter *dtp-version*
      (format nil "~D.~D alpha [~A,~A,~A]"
	      *dtp-major-version*
	      *dtp-minor-version*
	      *dtp-compilation-status*
	      *dtp-tracing-status*
	      *dtp-typing-status* ))
  (export '(*dtp-version*)) )


(eval-when (compile load eval)
  (format t "~&DTP version ~A~%" *dtp-version*) )

;;;----------------------------------------------------------------------------

(eval-when (compile load eval)
  (defparameter *dtp-directory*
      #-(or lucid allegro-v4.1) "dtp:src;"
      #+(and allegro-v4.1 customize-dfg) "~/DTP/Source/"
      #+(and allegro-v4.1 (not customize-dfg)) ""
      #+(and lucid customize-dfg) "/lisp/dtp/"
      #+(and lucid (not customize-dfg)) "" )
  (defparameter *dtp-logic-directory*
      #-(or lucid allegro-v4.1) "dtp:logic;"
      #+(and allegro-v4.1 customize-dfg) "~/DTP/Logic/"
      #+(and allegro-v4.1 (not customize-dfg)) ""
      #+(and lucid customize-dfg) "/lisp/dtp/"
      #+(and lucid (not customize-dfg)) "" )
  (export '(*dtp-directory* *dtp-logic-directory*)) )

(eval-when (compile load eval)
  (unless (find :dtp-defsystem *features*)
    (load (concatenate 'string *dtp-directory* "defsystem")) ))

;;;----------------------------------------------------------------------------

(dtp-ds:defsystem :dtp
    :source-pathname #-(or lucid allegro-v4.1) "dtp:src;"
    #+(and allegro-v4.1 customize-dfg) "~/DTP/Source/"
    #+(and allegro-v4.1 (not customize-dfg)) ""
    #+(and lucid customize-dfg) "/lisp/dtp/"
    #+(and lucid (not customize-dfg)) ""
    :binary-pathname #-(or lucid allegro-v4.1) "dtp:bin;"
    #+(and allegro-v4.1 customize-dfg) "~/DTP/Compiled/"
    #+(and allegro-v4.1 (not customize-dfg)) ""
    #+(and lucid customize-dfg) "/lisp/dtp/"
    #+(and lucid (not customize-dfg)) ""
    :components
    ((:module :dtp-data-structures
	      :source-pathname ""
	      :components
	      ((:file "types")
	       (:file "variables")
	       (:file "structures")
	       (:file "classes") ))
     (:module :dtp-extensions
	      :source-pathname ""
	      :components
	      ((:file "internals")) )
     (:module :dtp-defsystem
	      :source-pathname ""
	      :components
	      ((:file "defsystem") ))
     (:module :dtp-unification
	      :source-pathname ""
	      :components
	      ((:file "symbols")
	       (:file "bindings")
	       (:file "binding-dag")
	       (:file "match")
	       (:file "cnf") ))
     (:module :dtp-kb
	      :source-pathname ""
	      :components
	      ((:file "literals")
	       (:file "clauses")
	       (:file "labels")
	       (:file "answers")
	       (:file "database") )
	      :depends-on (:dtp-data-structures :dtp-extensions) )
     (:module :dtp-inference
	      :source-pathname ""
	      :components
	      ((:file "misc-inference")
	       (:file "terms")
	       (:file "subgoals")
	       (:file "conjunctions")
	       (:file "conjunct")
	       (:file "backtrack")
	       (:file "caching")
	       (:file "residue")
	       (:file "ordering")
	       (:file "prover") )
	      :depends-on (:dtp-data-structures :dtp-extensions) )
     #+dtp-trace
     (:module :dtp-view
	      :source-pathname ""
	      :components
	      ((:file "view")
	       (:file "dotify")
	       (:file "textify")
	       (:file "below") )
	      :depends-on (:dtp-data-structures :dtp-extensions) )
     (:module :dtp-misc
	      :source-pathname ""
	      :components
	      ((:file "hierarchy")
	       (:file "output")
	       (:file "file")
	       (:file "test")
	       (:file "epikit-dtp") )
	      :depends-on (:dtp-data-structures :dtp-extensions) )
     ))

;;;----------------------------------------------------------------------------

(defun compile-dtp ()
  (dtp-ds:operate-on-system :dtp 'compile :verbose t) )

(defun load-dtp ()
  (dtp-ds:operate-on-system :dtp 'load :verbose t) )

(eval-when (load eval)
  (load-dtp) )

;;;----------------------------------------------------------------------------
