;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Prover.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

(eval-when (compile load eval)
  (export
   '(prove prove-next-answer prove-all-remaining-answers) ))

;;;----------------------------------------------------------------------------
;;;
;;;	Public

;;;----------------------------------------------------------------------------

(defun prove (query &key (all-answers nil) (return-form nil))
  "Returns (1) bound query(ies), (2) label(s), (3) <answer(s)>, (4) <proof>"
  (let (*proof*)
    (setq *proof* (make-proof :query query :return-form return-form))
    (setf (proof-query-conjunctions *proof*)
      (mapcar #'list-to-conjunction (dnf query)) )
    (when *use-negated-goal*
      (let ((*goal-node-id-count* 0))
	(setf (proof-goal-nodes *proof*)
	  (mapcar #'make-goal-node (proof-query-conjunctions *proof*)) )))
    (setq *last-proof* *proof*)
    (cond
     (all-answers
      (prove-all-remaining-answers *proof*) )
     (t
      (prove-next-answer *proof*) ))
    ))

;;;----------------------------------------------------------------------------

(defun prove-next-answer (&optional (*proof* *proof*))
  "Returns (1) bound query, (2) label, (3) <answer> or :NOT-AN-ANSWER, and
(4) <proof>"
  (let ((*proof-line-count* 0)
	(*depth* 0)
	(*subgoal-map* nil) )
    (explode-answer (prove-next-answer-with-iteration *proof*)) ))

;;;----------------------------------------------------------------------------

(defun prove-all-remaining-answers (&optional (*proof* *proof*))
  "Returns (1) queries, (2) labels, (3) residues, (4) <answer>s, (5) <proof>"
  (let ((*proof-line-count* 0)
	(*depth* 0)
	(*subgoal-map* nil)
	answers bounds label-values residues )
    (setq answers (prove-all-remaining-answers-internal *proof*))
    (setq bounds (mapcar #'apply-answer answers))
    (setq label-values (mapcar #'extract-label answers))
    (setq residues (mapcar #'extract-residue answers))
    (values bounds label-values residues answers *proof*) ))

;;;----------------------------------------------------------------------------

(defmethod propogate (answer (proof proof))
  (add-to-end answer (proof-new-answers proof)) )

;;;----------------------------------------------------------------------------

(defun apply-answer (answer &optional (form (proof-return-form *proof*)))
  "Turn an answer structure into a user-readable form"
  #+dtp-types (declare (type answer answer))
  (unless form (setq form (proof-query *proof*)))
  (if (answer-ae-binding-lists answer)
      (cons 'or
	    (cons (plug form (answer-binding-list answer))
		  (mapcar #'(lambda (bl) (plug form bl))
			  (answer-ae-binding-lists answer) )))
    (plug form (answer-binding-list answer)) ))

;;;----------------------------------------------------------------------------
;;;
;;;	Private

;;;----------------------------------------------------------------------------

(defun prove-next-answer-with-iteration (&optional (*proof* *proof*))
  "Iterate subgoal and function cutoffs, proving next answer"
  (unless (or *use-subgoal-cutoffs* *use-function-cutoffs*)
    (return-from
	prove-next-answer-with-iteration
      (prove-next-answer-internal) ))
  (prove-next-answer-internal) )

;;;----------------------------------------------------------------------------

(defun prove-next-answer-internal (&optional (*proof* *proof*))
  "Return <answer> or :NOT-AN-ANSWER"
  #+dtp-trace
  (when (find :answers *trace*)
    (format *debug-io* "Looking for next answer of ~A~%"
	    (proof-query *proof*) ))
  (loop
      while (or (proof-new-answers *proof*)
		(proof-subgoal-agenda *proof*)
		(proof-query-conjunctions *proof*) )
      for next-answer = (pop (proof-new-answers *proof*))
      for next-subgoal = (agenda-best)
      for next-conjunction = (first (proof-query-conjunctions *proof*))
      do
	(cond
	 (next-answer
	  (add-to-end next-answer (proof-answers *proof*))
	  (return next-answer) )
	 (next-subgoal
	  (cond
	   ((null (slot-value next-subgoal 'conjuncts-to-propogate-to))
	    #+dtp-trace
	    (when (find :proofs *trace*)
	      (format *debug-io* "Discarding irrelevant subgoal ")
	      (print-literal-node
	       (slot-value next-subgoal 'literal) :s *debug-io* )
	      (format *debug-io* "~%") )
	    (agenda-remove next-subgoal) )
	   ((exhausted-p next-subgoal)
	    #+dtp-trace
	    (when (find :proofs *trace*)
	      (format *debug-io* "Discarding exhausted subgoal ")
	      (print-literal-node
	       (slot-value next-subgoal 'literal) :s *debug-io* )
	      (format *debug-io* "~%") )
	    (incf *depth*)
	    (terminate next-subgoal)
	    (decf *depth*)
	    (agenda-remove next-subgoal) )
	   (t
	    (expand next-subgoal) )))
	 (t
	  (if (exhausted-p next-conjunction)
	      (add-to-end
	       (pop (proof-query-conjunctions *proof*))
	       (proof-used-conjunctions *proof*) )
	    (expand next-conjunction) )))
      finally (return :not-an-answer) ))

;;;----------------------------------------------------------------------------

(defun prove-all-remaining-answers-internal (&optional (*proof* *proof*))
  "Return list of all remaining <answers>"
  (loop
      for next-answer = (prove-next-answer-with-iteration *proof*)
      until (eq next-answer :not-an-answer)
      collect next-answer into answers
      #+dtp-trace do
	#+dtp-trace
	(when (find :answers *trace*)
	  (format *debug-io* "Found answer ~A~2%" next-answer) )
      finally (return answers) ))

;;;----------------------------------------------------------------------------

(defun explode-answer (answer)
  "Returns (1) bound query, (2) label, (3) residue, (4) ANSWER, (5) *PROOF*"
  #+dtp-types (declare (type answer answer))
  (if (answer-p answer)
      (values (apply-answer answer) (extract-label answer)
	      (extract-residue answer) answer *proof* )
    (values nil nil nil :not-an-answer *proof*) ))

;;;----------------------------------------------------------------------------

(defun extract-label (answer)
  #+dtp-types (declare (type answer answer))
  (let ((label (answer-label answer)))
    (when label (label-value label)) ))

;;;----------------------------------------------------------------------------

(defun extract-residue (answer)
  #+dtp-types (declare (type answer answer))
  (mapcar #'literal-to-list (answer-residue answer)) )

;;;----------------------------------------------------------------------------

(defun make-goal-node (conjunction)
  "Return clause corresponding to negated conjunction"
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (with-slots (list) conjunction
    (loop
	for conjunct in list
	for literal = (slot-value conjunct 'literal)
	for new-literal = (copy-literal-node literal)
	do (setf (literal-negated-p new-literal)
	     (not (literal-negated-p new-literal)) )
	collect new-literal into literals
	finally
	  (return
	    (make-kb-node
	     :id (make-new-id "GOAL" (incf *goal-node-id-count*))
	     :clause (make-clause-node :literals literals)) ))
    ))

;;;----------------------------------------------------------------------------

(defun agenda-add (subgoal)
  "Add SUBGOAL to the proof agenda"
  (if (find subgoal (proof-subgoal-agenda *proof*))
      (progn
	#+dtp-trace
	(when (find :proofs *trace*)
	  (indent-line)
	  (format *debug-io* "Subgoal ")
	  (print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
	  (format *debug-io* " already on the proof agenda at position ")
	  (format *debug-io* "~D~%"
		  (position subgoal (proof-subgoal-agenda *proof*)) )))
    (progn
      #+dtp-trace
      (when (find :proofs *trace*)
	(indent-line)
	(format *debug-io* "Pushing ")
	(print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
	(format *debug-io* " on the proof agenda~%") )
      (push subgoal (proof-subgoal-agenda *proof*)) )
    ))

;;;----------------------------------------------------------------------------

(defun agenda-remove (subgoal)
  "Remove SUBGOAL from the proof agenda"
  #+dtp-trace
  (when (find :proofs *trace*)
    (indent-line)
    (format *debug-io* "Removing ")
    (print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
    (format *debug-io* " from the proof agenda~%") )
  (if (eq subgoal (first (proof-subgoal-agenda *proof*)))
      (pop (proof-subgoal-agenda *proof*))
    (progn
      #+dtp-trace
      (when (find :proofs *trace*)
	(indent-line)
	(format *debug-io* "Note: Subgoal position on agenda is ~A~%"
		(position subgoal (proof-subgoal-agenda *proof*)) ))
      (setf (proof-subgoal-agenda *proof*)
	(remove subgoal (proof-subgoal-agenda *proof*)) ))
    ))

;;;----------------------------------------------------------------------------
