;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Structures.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

;;;----------------------------------------------------------------------------

(defstruct (proof
	    (:print-function print-proof) )
  ;; Specified by user
  query
  (theory *theory*)
  (return-form nil)		; NIL => plug in to original query

  ;; Internal
  query-conjunctions
  subgoal-agenda
  (new-answers nil)		; Acquired by propogation, will move...
  (answers nil)			; ...to here once processed
  (subgoal-index (make-hash-table :test #'eq))
				; Map from relation -> sg list
  goal-nodes
  (pure-literal-nodes (make-hash-table :test #'eq))
				; Cache for kb nodes to ignore
  (assumables (mapcar #'list-to-literal *assumables*))
				; List of LITERAL-NODEs to match
  (consistency-check *consistency-check*)
  
  ;; Iteration and cutoffs
  (subgoal-depth-cutoff *initial-subgoal-depth*)
  (subgoal-depth-skip *subgoal-depth-skip*)
  (subgoal-maximum-depth *subgoal-maximum-depth*)
  (function-depth-cutoff *initial-function-depth*)
  (function-depth-skip *function-depth-skip*)
  (function-maximum-depth *function-maximum-depth*)
  
  ;; Tracing
  #+dtp-trace (used-conjunctions nil)
  #+dtp-trace (forward-inference-conjunctions nil) )

(defun print-proof (structure stream depth)
  #+dtp-types (declare (type proof structure))
  #+dtp-types (declare (type stream stream))
  (declare (ignore depth))
  (format stream "#<Proof of ~A with ~D answer~:P"
	  (proof-query structure) (length (proof-answers structure)) )
  (when (null (proof-query-conjunctions structure))
    (format stream " [Complete]") )
  (format stream ">") )

;;;----------------------------------------------------------------------------

(defstruct (answer
	    (:print-function print-answer) )
  (binding-list nil)
  (label nil)
  (residue nil)
  (ae-binding-lists nil)
  #+dtp-trace
  (justification nil)
  #+dtp-trace
  proof )

(defun print-answer (structure stream depth)
  #+dtp-types (declare (type answer structure))
  #+dtp-types (declare (type stream stream))
  (declare (ignore depth))
  (format stream "#<Answer")
  (if (answer-binding-list structure)
      (print-binding-list (answer-binding-list structure) :s stream)
    (format stream " TRUE") )
  (when (answer-label structure)
    (format stream " with label ~A"
	    (label-value (answer-label structure)) ))
  (when (answer-residue structure)
    (format stream " with residue")
    (dolist (literal (answer-residue structure))
      (format stream " ")
      (print-literal-node literal :s stream) ))
  (dolist (ae-bl (answer-ae-binding-lists structure))
    (format stream " or")
    (print-binding-list ae-bl :s stream) )
  (format stream ">") )

;;;----------------------------------------------------------------------------

(defstruct (reduction-answer
	    (:include answer)
	    (:conc-name ranswer-)
	    (:print-function print-reduction-answer) )
  "An answer derived via reduction, so full binding list must be kept"
  (reduced-subgoals nil)	; Must pass through these to become answer
  )

(defun print-reduction-answer (structure stream depth)
  #+dtp-types (declare (type reduction-answer structure))
  #+dtp-types (declare (type stream stream))
  (declare (ignore depth))
  (format stream "#<Reduction answer")
  (if (answer-binding-list structure)
      (print-binding-list (answer-binding-list structure) :s stream)
    (format stream " TRUE") )
  (when (answer-label structure)
    (format stream " with label ~A"
	    (label-value (answer-label structure)) ))
  (dolist (ae-bl (answer-ae-binding-lists structure))
    (format stream " or")
    (print-binding-list ae-bl :s stream) )
  (format stream ">") )

;;;----------------------------------------------------------------------------

(defstruct (kb-node
	    (:print-function print-kb-node) )
  id
  (clause nil) )

(defun print-kb-node (structure stream depth)
  #+dtp-types (declare (type kb-node structure))
  #+dtp-types (declare (type stream stream))
  (declare (ignore depth))
  (format stream "#<~A:~A>"
	  (kb-node-id structure) (kb-node-clause structure) ))

;;;----------------------------------------------------------------------------

(defstruct (justification))

(defstruct (l-justification	; Lookup
	    (:include justification)
	    (:conc-name l-just-) )
  id )

(defstruct (s-justification	; Subgoal
	    (:include justification)
	    (:conc-name s-just-) )
  conjunct-number
  subgoal
  answer
  t-bl )

(defstruct (c-justification	; Conjunction
	    (:include justification)
	    (:conc-name c-just-) )
  conjunction
  answer )

(defstruct (res-justification	; Residue
	    (:include justification)
	    (:conc-name res-just-) )
  assumable )

(defstruct (r-justification	; Reduction
	    (:include justification)
	    (:conc-name r-just-) )
  ancestor-subgoal
  leaf-subgoal )

;;;----------------------------------------------------------------------------
