;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Subgoals.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

;;;----------------------------------------------------------------------------
;;;
;;;	Public

;;;----------------------------------------------------------------------------

(defmethod expand ((subgoal dtp-subgoal))
  (let (answer)
    #+dtp-types (declare (type (or answer null) answer))
    (cond
     ((and *use-iap* (has-identical-ancestor subgoal))
      #+dtp-trace
      (when (find :proofs *trace*)
	(indent-line)
	(format *debug-io* "Subgoal ")
	(print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
	(format *debug-io* " pruned because of identical ancestor~%") )
      (agenda-remove subgoal)
      (terminate subgoal) )
     ((setq answer (expand-reduction subgoal))
      (propogate answer subgoal) )
     ((setq answer (expand-residue subgoal))
      (propogate answer subgoal) )
     (t
      (with-slots (inferences #+dtp-trace used-inferences) subgoal
	(when (eq inferences :uninitialized)
	  (setq inferences (norder-conjunctions (compute-inference subgoal))) )
	(if inferences
	    (let ((conjunction (first inferences)))
	      (with-slots (list binding-list label ae-binding-list) conjunction
		(cond
		 ((null list)
		  (pop inferences)
		  #+dtp-trace (add-to-end conjunction used-inferences)
		  (propogate
		   (make-answer
		    :binding-list binding-list :label label
		    :ae-binding-lists
		    (when ae-binding-list (list ae-binding-list))
		    #+dtp-trace :justification
		    #+dtp-trace
		    (make-l-justification :id (slot-value conjunction 'origin))
		    #+dtp-trace :proof #+dtp-trace *proof* )
		   subgoal ))
		 ((exhausted-p conjunction)
		  (pop inferences)
		  #+dtp-trace (add-to-end conjunction used-inferences)
		  (unless inferences
		    #+dtp-trace
		    (when (find :proofs *trace*)
		      (indent-line)
		      (format *debug-io* "Removing exhausted subgoal ")
		      (print-literal-node
		       (slot-value subgoal 'literal) :s *debug-io* )
		      (format *debug-io* "~%") )
		    ;; (terminate subgoal)
		    (agenda-remove subgoal)
		    ))
		 (t
		  (expand conjunction) ))
		))
	  (terminate subgoal) ))
      ))))

#+dtp-trace
(defmethod expand :around ((subgoal dtp-subgoal))
  (when (find :subgoals *trace*)
    (indent-line)
    (format *debug-io* "Expanding subgoal ")
    (print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
    (format *debug-io* " [~D]~%" (slot-value subgoal 'depth)) )
  (incf *depth*)
  (prog1
      (call-next-method)
    (decf *depth*) ))

;;;----------------------------------------------------------------------------

(defmethod exhausted-p ((subgoal dtp-subgoal))
  (null (slot-value subgoal 'inferences)) )

(defmethod virgin-subgoal-p ((subgoal dtp-subgoal))
  "True iff SUBGOAL has never been worked on"
  (and
   (eq (slot-value subgoal 'inferences) :uninitialized)
   (null (slot-value subgoal 'reduction-answers)) ))

;;;----------------------------------------------------------------------------

(defmethod propogate (answer (subgoal dtp-subgoal))
  (with-slots
      (answers reduction-answers conjuncts-to-propogate-to parent-subgoal)
      subgoal

    ;; If a reduction reached the uppermost goal, then convert to normal answer
    (when (and (typep answer 'reduction-answer)
	       (find subgoal (ranswer-reduced-subgoals answer)) )
      (setq answer (copy-reduction-answer answer))
      (setf (ranswer-reduced-subgoals answer)
	(remove subgoal (ranswer-reduced-subgoals answer)) )
      (unless (ranswer-reduced-subgoals answer)
	(let ((new-answer (make-answer)))
	  (setf (answer-binding-list new-answer) (answer-binding-list answer))
	  (setf (answer-label new-answer) (answer-label answer))
	  (setf (answer-residue new-answer) (answer-residue answer))
	  (setf (answer-ae-binding-lists new-answer)
	    (answer-ae-binding-lists answer) )
	  #+dtp-trace
	  (setf (answer-justification new-answer)
	    (answer-justification answer) )
	  #+dtp-trace (setf (answer-proof new-answer) *proof*)
	  (setq answer new-answer) )))
    
    (if (typep answer 'reduction-answer)
	(unless (find answer reduction-answers :test #'answer-equal-p)
	  (add-to-end answer reduction-answers)
	  (let (conjs-of-original-parent)
	    (setq conjs-of-original-parent
	      (remove-if-not
	       #'(lambda (conj)
		   (eq parent-subgoal
		       (slot-value
			(slot-value conj 'parent-conjunction)
			'parent-subgoal )))
	       conjuncts-to-propogate-to ))
	    (when conjs-of-original-parent
	      (dolist (conjunct conjs-of-original-parent)
		(propogate answer conjunct) ))))
      (progn
	(setq answer (nsimplify-binding-list (copy-answer answer) subgoal))
	(unless (find answer answers :test #'answer-equal-p)
	  (add-to-end answer answers)
	  (dolist (conjunct conjuncts-to-propogate-to)
	    (propogate answer conjunct) ))))
    ))

#+dtp-trace
(defmethod propogate :around (answer (subgoal dtp-subgoal))
  "For proof tracing"
  (when (find :proofs *trace*)
    (indent-line)
    (format *debug-io* "Propogating ~S to subgoal " answer)
    (print-literal-node (slot-value subgoal 'literal) :s *debug-io*)
    (format *debug-io* "~%") )
  (incf *depth*)
  (prog1
    (call-next-method)
    (decf *depth*) ))

;;;----------------------------------------------------------------------------

(defmethod terminate ((subgoal dtp-subgoal))
  "Let waiting conjuncts know that there are no more answers"
  (dolist (conjunct (slot-value subgoal 'conjuncts-to-propogate-to))
    (propogate :not-an-answer conjunct) ))

;;;----------------------------------------------------------------------------
;;;
;;;	Private

;;;----------------------------------------------------------------------------

(defun expand-reduction (subgoal)
  #+dtp-types (declare (type dtp-subgoal subgoal))
  (unless *use-reduction*
    (return-from expand-reduction nil) )
  (with-slots (remaining-ancestor-subgoals #+dtp-trace used-ancestors) subgoal
    (when (eq remaining-ancestor-subgoals :uninitialized)
      (setq remaining-ancestor-subgoals (subgoal-ancestors-of subgoal)) )
    (loop
	for ancestor = (pop remaining-ancestor-subgoals)
	while ancestor
	for answer = (reduction subgoal ancestor)
	do #+dtp-trace
	   (add-to-end ancestor used-ancestors)
	   (when answer
	     #+dtp-trace
	     (when (find :proofs *trace*)
	       (indent-line)
	       (format *debug-io* "Found ~A by reduction~%" answer)
	       (indent-line)
	       (format *debug-io* " with ~A~%" ancestor) )
	     (return answer) ))
    ))

;;;----------------------------------------------------------------------------

(defun expand-residue (subgoal)
  #+dtp-types (declare (type dtp-subgoal subgoal))
  (unless *use-residue*
    (return-from expand-residue nil) )
  (with-slots (literal assumables) subgoal
    (when (eq assumables :uninitialized)
      (setq assumables (proof-assumables *proof*)) )
    (loop
	for assumable = (pop assumables)
	while assumable
	for answer = (residue-answer subgoal assumable)
	when answer
	do #+dtp-trace
	   (when (find :proofs *trace*)
	     (indent-line)
	     (format *debug-io* "Assuming residue ")
	     (print-literal-node literal :s *debug-io*)
	     (format *debug-io* " to get ~A~%" answer) )
	   (return answer) )))

;;;----------------------------------------------------------------------------

(defun compute-inference (subgoal)
  "Returns list of conjunctions"
  #+dtp-types (declare (type dtp-subgoal subgoal))
  (unless *use-subgoal-inference*
    (return-from compute-inference nil) )
  (loop
      with sg-literal = (slot-value subgoal 'literal)
      with node-list-1 = (proof-goal-nodes *proof*)
      with node-list-2 =
	(remove-pure-literal-nodes
	 (active-theory-contents
	  (proof-theory *proof*)
	  (literal-relation (slot-value subgoal 'literal)) ))
      for resolving-with-goal-p = node-list-1
      for kb-node =
	(if node-list-1 (pop node-list-1) (pop node-list-2))
      while kb-node
      appending
	(loop
	    with c-bl =
	      (multiple-value-list
		  (clause-rename-all-variables (kb-node-clause kb-node)) )
	    with clause = (first c-bl)
	    with tbl = (second c-bl)
	    for literal in
	      (if *use-contrapositives*
		  (clause-literals clause)
		(list (first (clause-literals clause))) )
	    for mgu = (literal-mgu literal sg-literal)
	    when mgu
	    collect
	      (resolve mgu sg-literal literal clause subgoal
		       (kb-node-id kb-node)
		       (when resolving-with-goal-p tbl) ))
	))

;;;----------------------------------------------------------------------------

(defun resolve (mgu sg-lit kb-lit clause parent kb-parent ans-ext-bl)
  "Returns conjunction"
  #+dtp-types (declare (type binding-list mgu))
  #+dtp-types (declare (type literal-node sg-lit kb-lit))
  #+dtp-types (declare (type clause-node clause))
  #+dtp-types (declare (type dtp-subgoal parent))
  #+dtp-types (declare (type symbol kb-parent))
  #+dtp-types (declare (type binding-list ans-ext-bl))
  (setq ans-ext-bl (remove '(t . t) ans-ext-bl :test #'equal))
  (let ((sg-vars (find-vars (literal-terms sg-lit)))
	bl aebl conjunction )
    #+dtp-types (declare (type dtp-conjunction conjunction))
    (setq clause (copy-clause-node clause))
    (setf (clause-literals clause) (remove kb-lit (clause-literals clause)))
    (setq clause (clause-plug clause mgu))
    (nclause-flip-negations clause)
    (setq bl
      (remove-if-not
       #'(lambda (binding)
	   #+dtp-types (declare (type binding binding))
	   (find (binding-variable binding) sg-vars) )
       mgu ))
    (setq aebl (plug ans-ext-bl mgu))
    (unless (different-binding-lists bl aebl)
      (setq aebl nil) )
    (setq conjunction
      (make-instance 'dtp-conjunction
	:parent parent
	:origin kb-parent
	:binding-list bl
	:label (clause-label clause)
	:ae-binding-list aebl ))
    (setf (slot-value conjunction 'list)
      (mapcar
       #'(lambda (lit)
	   #+dtp-types (declare (type literal-node lit))
	   (make-instance 'dtp-conjunct :literal lit :parent conjunction) )
       (clause-literals clause) ))
    conjunction ))

;;;----------------------------------------------------------------------------

(defun subgoal-ancestors-of (subgoal)
  "List of all direct parent subgoals above SUBGOAL in the proof graph"
  #+dtp-types (declare (type (or dtp-subgoal null) subgoal))
  (when subgoal (subgoal-ancestors-of-internal subgoal nil nil)) )

(defun subgoal-ancestors-of-internal (subgoal ancestors-so-far include-me)
  #+dtp-types (declare (type dtp-subgoal subgoal))
  #+dtp-types (declare (type list ancestors-so-far))
  (if (find subgoal ancestors-so-far)
      ancestors-so-far
    (with-slots (parent-subgoal) subgoal
      (when include-me
	(push subgoal ancestors-so-far) )
      (if parent-subgoal
	  (subgoal-ancestors-of-internal parent-subgoal ancestors-so-far t)
	ancestors-so-far ))))

(defun subgoal-parent-of (object)
  #+dtp-types (declare (type dtp-object object))
  (case (type-of object)
    (dtp-conjunct
     (slot-value (slot-value object 'parent-conjunction) 'parent-subgoal) )
    (dtp-conjunction
     (slot-value object 'parent-subgoal) )
    (dtp-subgoal
     object )))

;;;----------------------------------------------------------------------------

(defun reduction (subgoal ancestor)
  (let ((lit1 (slot-value subgoal 'literal))
	(lit2 (slot-value ancestor 'literal))
	mgu )
    (setq mgu (literal-negated-pair-p lit1 lit2))
    (when mgu
      (make-reduction-answer
       :binding-list mgu
       #+dtp-trace :justification
       #+dtp-trace
       (make-r-justification
	:ancestor-subgoal ancestor
	:leaf-subgoal subgoal )
       :reduced-subgoals (list ancestor) ))
    ))

;;;----------------------------------------------------------------------------
;;;
;;;	Pure Literal Elimination

(defun remove-pure-literal-nodes (nodes)
  #+dtp-types (declare (type list nodes))
  (unless *use-pure-literal-elimination*
    (return-from remove-pure-literal-nodes nodes) )
  (loop
      for node
	  #-lucid of-type #-lucid kb-node
      in nodes
      unless (pure-literal-node-p node)
      collect node ))

(defun pure-literal-node-p (node)
  #+dtp-types (declare (type kb-node node))
  (let (pure-p)
    (setq pure-p
      (gethash (kb-node-id node) (proof-pure-literal-nodes *proof*) :unknown) )
    (when (eq pure-p :unknown)
      (setq pure-p
	(loop
	    named check-pure
	    for literal in (clause-literals (kb-node-clause node))
	    unless (can-find-matching-literal literal :except node)
	    do #+dtp-trace
	       (when (find :proofs *trace*)
		 (format *debug-io* "[Pure literal ")
		 (print-literal-node-as-logic literal :s *debug-io*)
		 (format *debug-io* " detected in ~A...removing~%~7T"
			 (kb-node-id node) )
		 (print-clause-node
		  (kb-node-clause node) :s *debug-io* :as-rule t )
		 (format t "~% for duration of proof]~%") )
	       (return-from check-pure t)
	    finally (return-from check-pure nil) ))
      (setf (gethash (kb-node-id node) (proof-pure-literal-nodes *proof*))
	pure-p ))
    pure-p ))

(defun can-find-matching-literal (literal &key (except nil))
  #+dtp-types (declare (type literal-node literal))
  #+dtp-types (declare (type kb-node except))
  (or
   ;; Matching in query or (active) database...
   (loop
       for kb-node in
	 (append
	  (proof-goal-nodes *proof*)
	  (remove
	   except
	   (active-theory-contents
	    (proof-theory *proof*) (literal-relation literal) )))
       do
	 (loop
	     with test-clause =
	       (clause-rename-all-variables (kb-node-clause kb-node))
	     for kb-literal in (clause-literals test-clause)
	     when (literal-negated-pair-p literal kb-literal)
	     do (return-from can-find-matching-literal kb-node) )
       finally (return nil) )
   ;; ...or assumptions
   (find literal (proof-assumables *proof*) :test #'literal-negated-pair-p)
   ))

;;;----------------------------------------------------------------------------

(defun has-identical-ancestor (subgoal)
  "T iff subgoal has an identical parent"
  #+dtp-types (declare (type dtp-subgoal subgoal))
  (loop
      with sg-literal = (slot-value subgoal 'literal)
      for ancestor in (subgoal-ancestors-of subgoal)
      for ans-literal = (slot-value ancestor 'literal)
      when (literal-equal-p sg-literal ans-literal)
      return t
      finally (return nil) ))

;;;----------------------------------------------------------------------------
