;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		Test.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)

(in-package "DTP")

(eval-when (compile load eval)
  (export
   '(test-dtp load-logic-samples show-proofs) ))

;;;----------------------------------------------------------------------------

(defun test-dtp (&key (reset t))
  "Resets theorem prover, loads and runs test suite"
  (when reset
    (reset-dtp)
    (load-logic-samples) )
  (let (testfile)
    (setq testfile (concatenate 'string *dtp-logic-directory* "dtp.test"))
    #-(or lucid allegro-v4.1)
    (setq testfile (translate-logical-pathname testfile))
    (with-open-file (tf testfile :direction :input)
      (loop
	  with *package* = *dtp-package*
	  with errors = 0
	  with tests = 0
	  initially
	    (format t "~&")
	    (setq *test-proofs* nil)
	  for sexp = (read tf nil nil)
	  until (null sexp)
	  for expected-answer = (read tf nil nil)
	  with answer
	  with result
	  do #+dtp-trace
	     (when (find :tests *trace*)
	       (if (eq expected-answer *test-ignore-answer*)
		   (format t "[~(~S~)]~%" sexp)
		 (format t "~(~S~)~%" sexp) ))
	     (setq result (multiple-value-list (eval sexp)))
	     (setq answer (first result))
	     (unless (eq expected-answer *test-ignore-answer*)
	       (push (fifth result) *test-proofs*) )
	     #+dtp-trace
	     (when (and (find :tests *trace*)
			(not (eq expected-answer *test-ignore-answer*)) )
	       (format t "-> ~S~%" answer) )
	     (unless (eq expected-answer *test-ignore-answer*)
	       (incf tests)
	       (unless (equal expected-answer answer)
		 (incf errors)
		 #+dtp-trace
		 (when (find :tests *trace*)
		   (format t "Error: Should have gotten~%   ~A~%"
			   expected-answer ))
		 ))
	  finally
	    (setq *test-proofs*
	      (remove-if-not #'proof-p (reverse *test-proofs*)) )
	    #+dtp-trace
	    (when (find :tests *trace*)
	      (format t "~D test~:P checked, ~D error~:P~%" tests errors) )
	    (return errors) ))))

;;;----------------------------------------------------------------------------

(defun load-logic-samples ()
  "Load the sample logical theories into DTP"
  (dolist (file *logic-samples*)
    (dolist (theory (dtp-load (string-downcase file)))
      (unless (find theory (included-active-theory-names file))
	(includes file theory) ))
    (includes 'logic-samples file) ))

;;;----------------------------------------------------------------------------

(defun show-proofs (&optional (proofs *test-proofs*))
  (format t "~&~D proofs~%" (length proofs))
  (dolist (proof proofs)
    (format t "Proof of ~A in theory ~A~%"
	    (proof-query proof) (proof-theory proof) )
    (loop
	with *proof* = proof
	for count from 1
	for answer in (proof-answers proof)
	do (format t "  ~2D: ~A~%" count (apply-answer answer)) )
    (format t "0-~D [Next Proof]: " (length (proof-answers proof)))
    (loop
	for line = (read-line)
	until (equal line "")
	for num = (read-from-string line)
	do (cond
	    ((not (integerp num))
	     (format t "Not a number~%") )
	    ((= num 0)
	     (show proof) )
	    ((> num (length (proof-answers proof)))
	     (format t "Out of range~%") )
	    (t
	     (show (nth (1- num) (proof-answers proof))) ))
	   (format t "0-~D [Next Proof]: " (length (proof-answers proof))) )
    ))

;;;----------------------------------------------------------------------------
