;;; -*- Mode: LISP; Syntax: Common-lisp; Package: DTP; Base: 10 -*-

;;;----------------------------------------------------------------------------
;;;
;;;	File		View.Lisp
;;;	System		Don's Theorem Prover
;;;
;;;	Written by	Don Geddis (Geddis@CS.Stanford.Edu)
;;;
;;;	Note		This entire file is conditional on #+dtp-trace

(in-package "DTP")

(eval-when (compile load eval)
  (export
   '(show) ))

;;;----------------------------------------------------------------------------
;;;
;;;	Public

;;;----------------------------------------------------------------------------

(defgeneric show (object))

(defmethod show ((object proof))
  (let ((*proof* object))
    (show-toplevel object) ))

(defmethod show ((object answer))
  (let ((*proof* (answer-proof object)))
    (show-toplevel object) ))

;;;----------------------------------------------------------------------------
;;;
;;;	Private

;;;----------------------------------------------------------------------------

(defvar *show-stream* nil "Output on file (for dot) or term (for text)")
(defvar *explored* nil "If we've seen this node before, don't explore below")
(defvar *id-assigned* nil "List of objects (id = position in list)")
(defvar *unique-id-assigned* nil "List of objects (negative id = pos in list)")

(defvar *debug-view* nil "Sends DOT output to *standard-io* instead")
(defvar *all-answers* nil)

;;;----------------------------------------------------------------------------

(defun show-toplevel (object)
  #+dtp-types (declare (type (or proof answer) object))
  (setq *explored* nil)
  (setq *id-assigned* nil)
  (setq *unique-id-assigned* nil)
  (if *graphic-display*
      (progn
	(with-open-file (f *dot-file* :direction :output :if-exists :supersede)
	  (setq *show-stream* f)
	  (when *debug-view*
	    (setq *show-stream* *standard-output*)
	    (setq *all-answers* nil) )
	  (dot-header object)
	  (output object)
	  (dot-footer object) )
	(if *debug-view*
	    (setq *all-answers* (reverse *all-answers*))
	  (progn
	    (when (probe-file *ps-file*)
	      (delete-file *ps-file*) )
	    #+allegro
	    (excl:run-shell-command
	     (format nil "~A -Tps ~A > ~A"
		     *dot-command* *dot-file* *ps-file* ))
	    #+allegro
	    (excl:run-shell-command
	     (format nil "~A~A ~A"
		     *ps-viewer-command*
		     (if *display-landscape* " -landscape" "")
		     *ps-file* )
	     :wait nil ))))
    (progn
      (setq *show-stream* *standard-output*)
      (let ((*depth* 0))
	(output object) ))))

;;;----------------------------------------------------------------------------

(defun output (object &rest rest &key (answer nil) &allow-other-keys)
  #+dtp-type (declare (type dtp-object object))
  (if answer
      (let ((my-id (dot-id-of object :unique t)))
	(when *debug-view* (push (cons object answer) *all-answers*))
	(if *graphic-display*
	    (apply #'dotify object :my-id my-id rest)
	  (apply #'textify object :my-id my-id rest) )
	(apply #'show-below object :my-id my-id rest) )
    (progn
      (if *graphic-display*
	  (apply #'dotify object rest)
	(apply #'textify object rest) )
      (apply #'show-below object rest) )))

;;;----------------------------------------------------------------------------
;;;
;;;	Dot headers and footers

(defgeneric dot-header (object))

(defun general-dot-header (&key (proof nil) (answer nil))
  (format *show-stream* "/* DTP ~A */~%" *dtp-version*)
  (when proof
    (format *show-stream* "/* Proof of ~A */~%" (proof-query proof)) )
  (when answer
    (format *show-stream* "/* Proof of ~A */~%" (apply-answer answer)) )
  (format *show-stream* "strict digraph proof {~%")
  (format *show-stream* "  page = \"8.5,11\" ;~%")
  (format *show-stream* "  margin = \".25,.25\" ;~%")
  (format *show-stream* "  ordering = out ;~%")
  (when *display-one-page*
    (format *show-stream* "  size = \"8,10.5\" ;~%")
    (format *show-stream* "  ratio = compress ;~%")
    (format *show-stream* "  center = true ;~%") )
  (when *display-landscape*
    (format *show-stream* "  orientation = landscape ;~%") )
  (format *show-stream* "  ranksep = .2 ;~%")
  (format *show-stream* "  nodesep = .1 ;~%")
  (format *show-stream* "  fontsize = 14 ;~%")
  (format *show-stream* "  node [fontsize=11,height=0,width=0] ;~%")
  (format *show-stream* "  edge [fontsize=9,weight=5] ;~%")
  )

(defmethod dot-header ((object proof))
  (general-dot-header :proof object)
  (when *display-title*
    (let ((*print-pretty* nil))
      (format *show-stream* "  label = \"\\nProof space of ~:(~A~)\\n"
	      (proof-query object) )
      (format *show-stream* "Theory ~:(~A~), ~D answer~:P, "
	      (proof-theory object) (length (proof-answers object)) )
      (if (or (proof-subgoal-agenda object)
	      (proof-query-conjunctions object) )
	  (format *show-stream* "In progress")
	(format *show-stream* "Complete") )
      (format *show-stream* "\" ;~%") ))
  )

(defmethod dot-header ((object answer))
  (general-dot-header :answer object)
  (when *display-title*
    (let ((*print-pretty* nil))
      (format *show-stream* "  label = \"\\nProof of ~:(~A~)\" ;~%"
	      (apply-answer object) )
      )))

(defgeneric dot-footer (object))

(defmethod dot-footer ((object proof))
  (format *show-stream* "}~%") )

(defmethod dot-footer ((object answer))
  (format *show-stream* "}~%") )

;;;----------------------------------------------------------------------------
;;;
;;;	Dot-Name-Of

(defgeneric dot-name-of (object &optional (answer nil)))

;;;----------------------------------------------------------------------------

(defmethod dot-name-of ((object proof) &optional (answer nil))
  (declare (ignore answer))
  (format nil "~:(~A~)" (proof-query object)) )

;;;----------------------------------------------------------------------------

(defmethod dot-name-of ((object answer) &optional (answer nil))
  (declare (ignore answer))
  (format nil "~:(~A~)" (apply-answer object)) )

;;;----------------------------------------------------------------------------

(defmethod dot-name-of ((object dtp-subgoal) &optional (answer nil))
  (with-output-to-string (s)
    (if (typep answer 'answer)
	(loop
	    with lit = (slot-value object 'literal)
	    for bl in (answer-ae-binding-lists answer)
	    collect (literal-plug lit bl) into lits
	    finally
	      (setq lits
		(cons (literal-plug lit (answer-binding-list answer)) lits) )
	      (setq lits (remove-duplicates lits :test #'literal-equal-p))
	      (loop
		  for each-lit in lits
		  for between = nil then t
		  do (when between (format s " or "))
		     (print-literal-node each-lit :s s) ))
      (print-literal-node (slot-value object 'literal) :s s) )))

;;;----------------------------------------------------------------------------

(defmethod dot-name-of ((object dtp-conjunction) &optional (answer nil))
  (if (slot-value object 'list)
      (with-output-to-string (s)
	(loop
	    with bl = (when (typep answer 'answer)
			(answer-binding-list answer) )
	    with conjuncts = (slot-value object 'list)
	    for count from 0
	    for conjunct = (pop conjuncts)
	    while conjunct
	    do (format s "<c~D> " count)
	       (print-literal-node
		(if bl
		    (literal-plug (slot-value conjunct 'literal) bl)
		  (slot-value conjunct 'literal) )
		:s s )
	       (when conjuncts (format s "|")) ))
    :lookup ))

;;;----------------------------------------------------------------------------

(defmethod dot-name-of ((object dtp-conjunct) &optional (answer nil))
  (declare (ignore answer))
  (with-output-to-string (s)
    (print-literal-node (slot-value object 'literal) :s s) ))

;;;----------------------------------------------------------------------------

(defun dot-id-of (object &key (unique nil))
  "Return integer ID...if UNIQUE, then every new call results in new ID"
  #+dtp-types (declare (type dtp-object object))
  (if unique
      (progn
	(add-to-end object *unique-id-assigned*)
	(- (length *unique-id-assigned*)) )
    (progn
      (add-new-to-end object *id-assigned*)
      (position object *id-assigned*) )))

(defun dot-id-to-object (id)
  (if (>= id 0)
      (nth id *id-assigned*)
    (nth (1- (- id)) *unique-id-assigned*) ))

(defun dot-node-name (id-or-object)
  (let ((id id-or-object))
    (unless (typep id-or-object 'integer)
      (setq id (dot-id-of id-or-object)) )
    (if (>= id 0)
	(format nil "n~D" id)
      (format nil "nu~D" (- id)) )))

;;;----------------------------------------------------------------------------

(defun binding-label-between (subgoal conjunction &optional (answer nil))
  #+dtp-types (declare (type dtp-subgoal subgoal))
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let (binding-list answer-vars)
    (setq binding-list (bl-from-inference subgoal conjunction))
    (when answer
      (setq answer-vars
	(mapcar #'binding-variable (answer-binding-list answer)) )
      (setq binding-list
	(remove-if #'(lambda (var) (find var answer-vars)) binding-list) ))
    (if binding-list
	(dot-name-of-bl binding-list)
      (when (setq binding-list (slot-value conjunction 'ae-binding-list))
	(dot-name-of-bl binding-list :disjunction t) ))
    ))

(defun bl-from-inference (subgoal conjunction)
  "Return binding from inference which affects subgoal"
  #+dtp-types (declare (type dtp-subgoal subgoal))
  #+dtp-types (declare (type dtp-conjunction conjunction))
  (let ((good-vars (literal-vars-in (slot-value subgoal 'literal))))
    (remove-if-not
     #'(lambda (binding)
	 #+dtp-types (declare (type binding binding))
	 (find (binding-variable binding) good-vars) )
     (slot-value conjunction 'binding-list) )))

(defun dot-name-of-bl (binding-list &key (disjunction nil))
  "Takes a binding list, returns a DOT label (string)"
  (with-output-to-string (s)
    (loop
	for binding-pair = (pop binding-list)
	while binding-pair
	unless (eq (binding-variable binding-pair) 't)
	do (when disjunction (format s "Or "))
	   (format s "~(~A~)->"
		   (variable-to-string (binding-variable binding-pair)) )
	   (term-to-string (binding-value binding-pair) s)
	   (when binding-list
	     (format s "\\n") ))
    ))

;;;----------------------------------------------------------------------------

(defun hidden-conjunction (conj)
  #+dtp-types (declare (type dtp-conjunction conj))
  (list-of-length-one-p (slot-value conj 'list)) )

(defun hidden-subgoal (parent-id)
  #+dtp-types (declare (type (or null integer) parent-id))
  (when parent-id
    (typep (dot-id-to-object parent-id) 'dtp-conjunction) ))

;;;----------------------------------------------------------------------------
